import os
import regex
import string
import logging
import csv
import traceback
from functools import partial

from collections import Counter
from unicodedata import normalize
from typing import List, Optional, Union
from bs4 import BeautifulSoup

from core.config import Settings
from musaddiquehussainlabs.text_preprocessing import emo_unicode, slang_text, contractions
from handlers.custom_exceptions import CustomJSONError
from handlers.output_generator import generate_output
from core.constants import constants, preprocess_operations


import nltk

from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize, PunktSentenceTokenizer
from nltk.stem import PorterStemmer, SnowballStemmer, LancasterStemmer, WordNetLemmatizer


def to_lower(input_text: str) -> str:
    """ Convert input text to lower case """
    return input_text.lower()


def to_upper(input_text: str) -> str:
    """ Convert input text to upper case """
    return input_text.upper()


def expand_contraction(input_text: str) -> str:
    """ Expand contractions in input text """
    return contractions.fix(input_text)    


def remove_number(input_text: str) -> str:
    """ Remove number in the input text """
    processed_text = regex.sub('\d+', '', input_text)
    return processed_text


def remove_itemized_bullet_and_numbering(input_text: str) -> str:
    """ Remove bullets or numbering in itemized input """
    processed_text = regex.sub('[(\s][0-9a-zA-Z][.)]\s+|[(\s][ivxIVX]+[.)]\s+', ' ', input_text)
    return processed_text


def remove_url(input_text: str) -> str:
    """ Remove url in the input text or ('https?://\S+|www\.\S+')"""
    return regex.sub('(www|http)\S+', '', input_text)


def remove_punctuation(input_text: str, punctuations: Optional[str] = None) -> str:
    """
    Removes all punctuations from a string, as defined by string.punctuation or a custom list.
    For reference, Python's string.punctuation is equivalent to '!"#$%&\'()*+,-./:;<=>?@[\\]^_{|}~'
    """
    if punctuations is None:
        punctuations = string.punctuation
    processed_text = input_text.translate(str.maketrans('', '', punctuations))
    return processed_text


def remove_special_character(input_text: str, special_characters: Optional[str] = None) -> str:
    """ Removes special characters """
    if special_characters is None:
        # TODO: add more special characters
        special_characters = 'å¼«¥ª°©ð±§µæ¹¢³¿®ä£'
    processed_text = input_text.translate(str.maketrans('', '', special_characters))
    return processed_text


def keep_alpha_numeric(input_text: str) -> str:
    """ Remove any character except alphanumeric characters """
    return ''.join(c for c in input_text if c.isalnum())


def remove_whitespace(input_text: str, remove_duplicate_whitespace: bool = True) -> str:
    """ Removes leading, trailing, and (optionally) duplicated whitespace """
    if remove_duplicate_whitespace:
        return ' '.join(regex.split('\s+', input_text.strip(), flags=regex.UNICODE))
    return input_text.strip()


def normalize_unicode(input_text: str) -> str:
    """ Normalize unicode data to remove umlauts, and accents, etc. """
    processed_tokens = normalize('NFKD', input_text).encode('ASCII', 'ignore').decode('utf8')
    return processed_tokens


def remove_stopword(input_text_or_list: Union[str, List[str]], stop_words: Optional[set] = None) -> List[str]:
    """ Remove stop words """

    if stop_words is None:
        stop_words = set(stopwords.words('english'))
    if isinstance(stop_words, list):
        stop_words = set(stop_words)
    if isinstance(input_text_or_list, str):
        tokens = word_tokenize(input_text_or_list)
        processed_tokens = [token for token in tokens if token not in stop_words]
    else:
        processed_tokens = [token for token in input_text_or_list
                            if (token not in stop_words and token is not None and len(token) > 0)]
    return processed_tokens


def remove_freqwords(input_text_or_list: Union[str, List[str]], freq_words: Optional[set] = None) -> List[str]:
    """ Remove freq words """

    cnt = Counter()

    if freq_words is None:
        freq_words = set([w for (w, wc) in cnt.most_common(10)])
    if isinstance(freq_words, list):
        freq_words = set(freq_words)
    if isinstance(input_text_or_list, str):
        tokens = word_tokenize(input_text_or_list)
        processed_tokens = [token for token in tokens if token not in freq_words]
    else:
        processed_tokens = [token for token in input_text_or_list
                            if (token not in freq_words and token is not None and len(token) > 0)]
    return processed_tokens


def remove_rarewords(input_text_or_list: Union[str, List[str]], rare_words: Optional[set] = None) -> List[str]:
    """ Remove freq words """

    cnt = Counter()
    n_rare_words = 10

    if rare_words is None:
        rare_words = set([w for (w, wc) in cnt.most_common()[:-n_rare_words-1:-1]])
    if isinstance(rare_words, list):
        rare_words = set(rare_words)
    if isinstance(input_text_or_list, str):
        tokens = word_tokenize(input_text_or_list)
        processed_tokens = [token for token in tokens if token not in rare_words]
    else:
        processed_tokens = [token for token in input_text_or_list
                            if (token not in rare_words and token is not None and len(token) > 0)]
    return processed_tokens


def remove_email(input_text: str) -> str:
    """ Remove email in the input text """
    regex_pattern = '[a-z0-9._%+-]+@[a-z0-9.-]+\.[a-z]{2,}'
    return regex.sub(regex_pattern, '', input_text)


def remove_phone_number(input_text: str) -> str:
    """ Remove phone number in the input text """
    regex_pattern = '(?:\+?(\d{1,3}))?[-. (]*(\d{3})[-. )]*(\d{3})[-. ]*(\d{4})(?: *x(\d+))?'
    return regex.sub(regex_pattern, '', input_text)


def remove_ssn(input_text: str) -> str:
    """ Remove social security number in the input text """
    regex_pattern = '(?!219-09-9999|078-05-1120)(?!666|000|9\d{2})\d{3}-(?!00)\d{2}-(?!0{4})\d{4}|(' \
                    '?!219099999|078051120)(?!666|000|9\d{2})\d{3}(?!00)\d{2}(?!0{4})\d{4}'
    return regex.sub(regex_pattern, '', input_text)


def remove_credit_card_number(input_text: str) -> str:
    """ Remove credit card number in the input text """
    regex_pattern = '(4[0-9]{12}(?:[0-9]{3})?|(?:5[1-5][0-9]{2}|222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][' \
                    '0-9]|2720)[0-9]{12}|3[47][0-9]{13}|3(?:0[0-5]|[68][0-9])[0-9]{11}|6(?:011|5[0-9]{2})[0-9]{12}|(' \
                    '?:2131|1800|35\d{3})\d{11})'
    return regex.sub(regex_pattern, '', input_text)


def remove_emoji(input_text: str) -> str:
    """ Remove Emoji in the input text """
    regex_pattern = regex.compile("["
                           u"\U0001F600-\U0001F64F"  # emoticons
                           u"\U0001F300-\U0001F5FF"  # symbols & pictographs
                           u"\U0001F680-\U0001F6FF"  # transport & map symbols
                           u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                           u"\U00002702-\U000027B0"
                           u"\U000024C2-\U0001F251"
                           "]+", flags=regex.UNICODE)
    return regex_pattern.sub(r'', input_text)


def remove_emoticons(input_text: str) -> str:
    """ Remove emoticons in the input text """
    regex_pattern = regex.compile(u'(' + u'|'.join(k for k in emo_unicode.EMOTICONS) + u')')
    return regex_pattern.sub(r'', input_text)


def convert_emoticons_to_words(input_text: str) -> str:
    """ Convert emoticons to words in the input text """
    for emot in emo_unicode.EMOTICONS:
        input_text = regex.sub(u'('+emot+')', "_".join(emo_unicode.EMOTICONS[emot].replace(",","").split()), input_text)

    return input_text


def convert_emojis_to_words(input_text: str) -> str:
    """ Convert emojis to words in the input text """
    for emot in emo_unicode.UNICODE_EMO:
        input_text = regex.sub(r'('+emot+')', "_".join(emo_unicode.UNICODE_EMO[emot].replace(",","").replace(":","").split()), input_text)

    return input_text


def remove_html(input_text: str) -> str:
    """ Remove html/xml in the input text """
    for emot in emo_unicode.UNICODE_EMO:
        return BeautifulSoup(input_text, "lxml").text


def chat_words_conversion(input_text: str) -> str:
    """ Chat Words Conversion in the input text """
    chat_words_map_dict = {}
    chat_words_list = []
    for line in slang_text.chat_words_str.split("\n"):
        if line != "":
            cw = line.split("=")[0]
            cw_expanded = line.split("=")[1]
            chat_words_list.append(cw)
            chat_words_map_dict[cw] = cw_expanded
    chat_words_list = set(chat_words_list)

    new_text = []
    for w in input_text.split():
        if w.upper() in chat_words_list:
            new_text.append(chat_words_map_dict[w.upper()])
        else:
            new_text.append(w)
    return " ".join(new_text)

def tokenize_word(input_text: str) -> List[str]:
    """ Converts a text into a list of word tokens """
    if input_text is None or len(input_text) == 0:
        return []
    return word_tokenize(input_text)

def tokenize_sentence(input_text: str) -> List[str]:
    """ Converts a text into a list of sentence tokens """
    if input_text is None or len(input_text) == 0:
        return []
    tokenizer = PunktSentenceTokenizer()
    return tokenizer.tokenize(input_text)


def stem_word(input_text_or_list: Union[str, List[str]],
              stemmer: Optional[Union[PorterStemmer, SnowballStemmer, LancasterStemmer]] = None
              ) -> List[str]:
    """ Stem each token in a text """
    if stemmer is None:
        stemmer = PorterStemmer()
    if isinstance(input_text_or_list, str):
        tokens = word_tokenize(input_text_or_list)
        processed_tokens = [stemmer.stem(token) for token in tokens]
    else:
        processed_tokens = [stemmer.stem(token) for token in input_text_or_list if token is not None and len(token) > 0]
    return processed_tokens


def lemmatize_word(input_text_or_list: Union[str, List[str]],
                   lemmatizer: Optional[WordNetLemmatizer] = None
                   ) -> List[str]:
    """ Lemmatize each token in a text by finding its base form """
    if lemmatizer is None:
        lemmatizer = WordNetLemmatizer()
    if isinstance(input_text_or_list, str):
        tokens = word_tokenize(input_text_or_list)
        processed_tokens = [lemmatizer.lemmatize(token) for token in tokens]
    else:
        processed_tokens = [lemmatizer.lemmatize(token)
                            for token in input_text_or_list if token is not None and len(token) > 0]
    return processed_tokens

def substitute_token(token_list: List[str], sub_dict: Optional[dict] = None) -> List[str]:
    """ Substitute each token by another token, e.g., 'vs' -> 'versus' """
    # TODO: add more custom substitutions in the csv file specified by _CUSTOM_SUB_CSV_FILE_PATH
    if token_list is None or len(token_list) == 0:
        return []
    if sub_dict is None:
        with open(Settings._CUSTOM_SUB_CSV_FILE_PATH, 'r') as f:
            csv_file = csv.reader(f)
            sub_dict = dict(csv_file)
    processed_tokens = list()
    for token in token_list:
        if token in sub_dict:
            processed_tokens.append(sub_dict[token])
        else:
            processed_tokens.append(token)
    return processed_tokens


# Create a dictionary mapping operation names to functions
operations_dict = {
    preprocess_operations.to_lower: to_lower,
    preprocess_operations.to_upper: to_upper,
    preprocess_operations.remove_number: remove_number,
    preprocess_operations.remove_punctuation: remove_punctuation,
    preprocess_operations.remove_stopword: remove_stopword,
    preprocess_operations.remove_itemized_bullet_and_numbering: remove_itemized_bullet_and_numbering,
    preprocess_operations.remove_url: remove_url,
    preprocess_operations.remove_special_character: remove_special_character,
    preprocess_operations.keep_alpha_numeric: keep_alpha_numeric,
    preprocess_operations.remove_whitespace: remove_whitespace,
    preprocess_operations.normalize_unicode: normalize_unicode,
    preprocess_operations.remove_freqwords: remove_freqwords,
    preprocess_operations.remove_rarewords: remove_rarewords,
    preprocess_operations.remove_email: remove_email,
    preprocess_operations.remove_phone_number: remove_phone_number,
    preprocess_operations.remove_ssn: remove_ssn,
    preprocess_operations.remove_credit_card_number: remove_credit_card_number,
    preprocess_operations.remove_emoji: remove_emoji,
    preprocess_operations.remove_emoticons: remove_emoticons,
    preprocess_operations.convert_emoticons_to_words: convert_emoticons_to_words,
    preprocess_operations.convert_emojis_to_words: convert_emojis_to_words,
    preprocess_operations.remove_html: remove_html,
    preprocess_operations.chat_words_conversion: chat_words_conversion,
    preprocess_operations.expand_contraction: expand_contraction,
    preprocess_operations.tokenize_word: tokenize_word,
    preprocess_operations.tokenize_sentence: tokenize_sentence,
    preprocess_operations.stem_word: stem_word,
    preprocess_operations.lemmatize_word: lemmatize_word,
    preprocess_operations.substitute_token: substitute_token
}




from core.config import Settings
from services.spacy_extractor import SpacyExtractor
from handlers.custom_exceptions import CustomJSONError
from handlers.output_generator import generate_output
from core.constants import constants
import spacy
import spacy.cli
import logging
import traceback

logger = logging.getLogger(__name__)
if not logger.handlers:
    logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(Settings.LOG_FILE_PATH)
    file_handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter(Settings.LOG_FILE_FORMAT)
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

# Please uncomment/comment below line if you wish to stop/start logging
# logger.disabled = True

try:
    nlp = spacy.load(Settings.SPACY_MODEL)
    extractor = SpacyExtractor(nlp)
except OSError:
    # If model is not available, download and install it
    spacy.cli.download(Settings.SPACY_MODEL)

class nlp:
    """NLP tasks"""

    def __init__(self, component_type: str = None) -> None:
        self.settings = Settings()
        self.component_type = component_type

    def load(self, component_type: str):
        try:
            logger.info("loading component_type: " + component_type +  " started...")

            nlp = spacy.load(Settings.SPACY_MODEL)
            
            doc = nlp("Apple is looking at buying U.K. startup for $1 billion")
            for token in doc:
                print(token.text)
            
        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message="Unexpected error occurred. Check logs for details.")

    @staticmethod
    def predict(component_type: str, input_text: str):
        try:
            
            if component_type is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_COMPONENT_TYPE_NONE)
            
            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
            
            if not isinstance(component_type, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_COMPONENT_TYPE_STRING)
            
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not component_type.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_COMPONENT_TYPE_EMPTY)   

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)           

            predicted_result = extractor.predict_nlp_component(component_type, input_text)
            
            if predicted_result is not None and bool(predicted_result):
                return predicted_result
            return generate_output(True, data=constants.COMPONENT_TYPE_UNSUPPORTED)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)

    def train(self):
        try:
            logger.info("data preprocessing started...")
            
        except Exception as ex:
            logger.error(str(ex))

    def PretrainedPipeline(self):
        try:
            logger.info("data preprocessing started...")
            
        except Exception as ex:
            logger.error(str(ex))




import json

def generate_output(succeeded, data=None, error_code=None, message=None):
    output = {"succeeded": succeeded}

    if succeeded:
        output["data"] = data
    else:
        output["error"] = {"error_code": error_code, "message": message}

    return json.dumps(output)



from pydantic_settings import BaseSettings

class AppConstants(BaseSettings):
    COMPONENT_TYPE_TOKENIZE: str = "tokenize"
    COMPONENT_TYPE_POS: str = "pos"
    COMPONENT_TYPE_LEMMATIZER: str = "lemma"
    COMPONENT_TYPE_MORPHOLOGY: str = "morphology"
    COMPONENT_TYPE_DEPENDENCY_PARSER: str = "dep"
    COMPONENT_TYPE_NER: str = "ner"
    COMPONENT_TYPE_NORMALIZERS: str = "norm"
    COMPONENT_TYPE_UNSUPPORTED: str = "Unsupported Component Type"
    ERROR_CODE_500: str = "500"
    ERROR_CODE_400: str = "400"
    UNEXPECTED_ERROR: str = "Unexpected error occurred. Check logs for details."
    VALIDATE_COMPONENT_TYPE_NONE: str = "Component type parameter should not be None."
    VALIDATE_INPUT_TEXT_NONE: str = "Input text parameter should not be None."
    VALIDATE_COMPONENT_TYPE_STRING: str = "Component type parameter should be a string."
    VALIDATE_INPUT_TEXT_STRING: str = "Input text parameter should be a string."
    VALIDATE_COMPONENT_TYPE_EMPTY: str = "Component type should not be empty or contain only whitespace."
    VALIDATE_INPUT_TEXT_EMPTY: str = "Input text parameter should not be empty or contain only whitespace."
    VALIDATE_INPUT_TEXT_EMPTY_NONE: str = "No appropriate positional argument is provide."
    TEXTPREPROCESSING_OPERATIONS_EMPTY_NONE: str = "Operations should be a non-empty list of strings"
    TEXTPREPROCESSING_INVALID_OPERATIONS: str = "Invalid operation detected: {0}. Only predefined operations are allowed."
    TEXTPREPROCESSING_OPERATIONS_NOT_CALLABLE: str = "Operation '{0}' not found or not callable."

constants =    AppConstants()

class TextPreprocessingConstants(BaseSettings):

    to_lower: str =  'to_lower'
    to_upper: str = 'to_upper'
    remove_number: str = 'remove_number'
    remove_punctuation: str =  'remove_punctuation'
    remove_stopword: str = 'remove_stopword'
    remove_itemized_bullet_and_numbering: str = 'remove_itemized_bullet_and_numbering'
    remove_url: str = 'remove_url'
    remove_special_character: str = 'remove_special_character'
    keep_alpha_numeric: str = 'keep_alpha_numeric'
    remove_whitespace: str = 'remove_whitespace'
    normalize_unicode: str = 'normalize_unicode'
    remove_freqwords: str = 'remove_freqwords'
    remove_rarewords: str = 'remove_rarewords'
    remove_email: str = 'remove_email'
    remove_phone_number: str = 'remove_phone_number'
    remove_ssn: str = 'remove_ssn'
    remove_credit_card_number: str = 'remove_credit_card_number'
    remove_emoji: str = 'remove_emoji'
    remove_emoticons: str = 'remove_emoticons'
    convert_emoticons_to_words: str = 'convert_emoticons_to_words'
    convert_emojis_to_words: str = 'convert_emojis_to_words'
    remove_html: str = 'remove_html'
    chat_words_conversion: str = 'chat_words_conversion'
    expand_contraction: str = 'expand_contraction'
    tokenize_word: str = 'tokenize_word'
    tokenize_sentence: str = 'tokenize_sentence'
    stem_word: str = 'stem_word'
    lemmatize_word: str = 'lemmatize_word'
    substitute_token: str = 'substitute_token'

preprocess_operations =    TextPreprocessingConstants()


from itertools import product
from textsearch import TextSearch

import json
import pkgutil

json_open = pkgutil.get_data("data", "raw/contractions_dict.json")
contractions_dict = json.loads(json_open.decode("utf-8"))

json_open = pkgutil.get_data("data", "raw/leftovers_dict.json")
leftovers_dict = json.loads(json_open.decode("utf-8"))

json_open = pkgutil.get_data("data", "raw/slang_dict.json")
slang_dict = json.loads(json_open.decode("utf-8"))

for month in [
    "january",
    "february",
    "march",
    "april",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december",
]:
    contractions_dict[month[:3] + "."] = month

contractions_dict.update({k.replace("'", "’"): v for k, v in contractions_dict.items()})

leftovers_dict.update({k.replace("'", "’"): v for k, v in leftovers_dict.items()})

safety_keys = set(
    ["he's", "he'll", "we'll", "we'd", "it's", "i'd", "we'd", "we're", "i'll", "who're", "o'"]
)

def get_combinations(tokens, joiners):
    combs = []
    combs.append(tokens)
    results = []
    for option in combs:
        option = [[x] for x in option]
        option = intersperse(option, joiners)
        for c in product(*option):
            results.append("".join(c))
    return results


def intersperse(lst, item):
    result = [item] * (len(lst) * 2 - 1)
    result[0::2] = lst
    return result


# have to create all the possibilities of ' and nothing
unsafe_dict = {}
for k, v in contractions_dict.items():  # contractions_dict.items():
    if k.lower() in safety_keys:
        continue
    if "'" not in k:
        continue
    tokens = k.split("'")
    for comb in get_combinations(tokens, ["", "'"]):
        unsafe_dict[comb] = v

slang_dict.update(unsafe_dict)

ts_leftovers = TextSearch("insensitive", "norm")
ts_leftovers.add(contractions_dict)
ts_leftovers.add(leftovers_dict)

ts_leftovers_slang = TextSearch("insensitive", "norm")
ts_leftovers_slang.add(contractions_dict)
ts_leftovers_slang.add(leftovers_dict)
ts_leftovers_slang.add(slang_dict)

ts_slang = TextSearch("insensitive", "norm")
ts_slang.add(contractions_dict)
ts_slang.add(slang_dict)

ts_basic = TextSearch("insensitive", "norm")
ts_basic.add(contractions_dict)

ts_view_window = TextSearch("insensitive", "object")
ts_view_window.add(list(contractions_dict.keys()))
ts_view_window.add(list(leftovers_dict.keys()))
ts_view_window.add(list(slang_dict.keys()))

replacers = {
    (True, False): ts_leftovers,
    (True, True): ts_leftovers_slang,
    (False, True): ts_slang,
    (False, False): ts_basic,
}


def fix(s, leftovers=True, slang=True):
    ts = replacers[(leftovers, slang)]
    return ts.replace(s)


def add(key, value):
    for ts in replacers.values():
        ts.add(key, value)


def preview(text, flank):
    """
    Return all contractions and their location before fix for manual check. Also provide a viewing window to quickly
    preview the contractions in the text.
    :param text: texture.
    :param flank: int number, control the size of the preview window. The window would be "flank-contraction-flank".
    :return: preview_items, a list includes all matched contractions and their locations.
    """
    try:
        int(flank)
    except Exception as e:
        print(e)
        raise Exception("Argument flank must be integer!")
    ts = ts_view_window
    results = ts.findall(text)
    preview_items = []
    for result in results:
        window_start = result.start - flank
        window_end = result.end + flank
        if window_start < 0:
            window_start = 0
        if window_end > len(text):
            window_end = len(text)
        preview_items.append({"match": result.match, "start": result.start, "end": result.end,
                             "viewing_window": text[window_start: window_end]})
    return preview_items


from pydantic_settings import BaseSettings

class prompts(BaseSettings):
    
    language_detection_template: str = """Detect the language in a given piece of text: {context}
        The detected language is:"""
    
    sentiment_analysis_template: str = """Analyze the sentiment of the following text as positive, neutral, negative and mixed and provide respective confidence_scores.\n{format_instructions}\n{question}"""
    keyphrase_extraction_template: str = """Extract key phrase from give text: {context}.\n{format_instructions}"""

prompt_templates =    prompts()


import datetime
import json
import typing
import logging
import traceback

import nltk
from nltk.tokenize import BlanklineTokenizer
import spacy

from core.config import Settings
from services.spacy_extractor import SpacyExtractor
from services.langchain_framework import LangchainFramework
from handlers.custom_exceptions import CustomJSONError
from handlers.output_generator import generate_output
from core.constants import constants

logger = logging.getLogger(__name__)
if not logger.handlers:
    logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(Settings.LOG_FILE_PATH)
    file_handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter(Settings.LOG_FILE_FORMAT)
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

# Please uncomment/comment below line if you wish to stop/start logging
# logger.disabled = True
    
try:
    nlp = spacy.load(Settings.SPACY_MODEL)
except OSError:
    # If model is not available, download and install it
    spacy.cli.download(Settings.SPACY_MODEL)

extractor = SpacyExtractor(nlp)
langchain = LangchainFramework()

class DocumentAnalysis:
    """Document Analysis class."""

    # pylint: disable=too-many-arguments
    def __init__(self) -> None:
        """Initialisation.
        Args:
            
        """

    def full_analysis(self, 
                      input_text: str
                      ) -> None:
        """analyze documents and return a dictionary containing the various parts of\
        the documents broken down into key-value pairs.
        Returns:
          dict: A dictionary with the content of the documents and broken down into
                key-value pairs.
        """
        try:

            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            self.input_text = input_text

            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure

            report_struc['language'] = langchain.language_detection(input_text)
            report_struc['linguistic_analysis'] = extractor.deep_linguistic_analysis(input_text)
            report_struc['keyphrases'] = langchain.keyphrase_extraction(input_text)
            report_struc['ner'] = extractor.entity_recognition(input_text)
            report_struc['sentiment'] = langchain.sentiment_analysis(input_text)
            report_struc['pii'] = langchain.pii_anonymization(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)

    def language_detection(self, input_text: str) -> typing.Dict[str, typing.Any]:
        try:

            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure
            report_struc['language'] = langchain.language_detection(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)

    def deep_linguistic_analysis(self, input_text: str) -> typing.Dict[str, typing.Any]:
        try:
            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure
            report_struc['disambiguaton'] = extractor.deep_linguistic_analysis(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)
    
    def keyphrase_extraction(self, input_text: str) -> typing.Dict[str, typing.Any]:
        try:
            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure
            report_struc['relevants'] = langchain.keyphrase_extraction(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)
    
    def entity_recognition(self, input_text: str) -> typing.Dict[str, typing.Any]:
        try:
            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure
            report_struc['ner'] = extractor.entity_recognition(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)
    
    def sentiment_analysis(self, input_text: str) -> typing.Dict[str, typing.Any]:
        try:
            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure
            report_struc['sentiment'] = langchain.sentiment_analysis(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)
    
    def pii_anonymization(self, input_text: str) -> typing.Dict[str, typing.Any]:
        try:
            if input_text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
            if not isinstance(input_text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

            if not input_text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
            
            input_text = "\n".join(input_text.splitlines())
            
            report_struc: typing.Dict[str, typing.Any] = {}  # Final structure
            report_struc['pii'] = langchain.pii_anonymization(input_text)

            return generate_output(True, data=report_struc)
        
        except CustomJSONError as e:
            logger.error(f"Custom JSON Error: {e.to_json()}")
            return e.to_json()

        except Exception as e:
            error_message = traceback.format_exc()
            logger.error(f"Unexpected error: {error_message}")
            return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)




# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing
from typing import Dict, List
import spacy
from spacy.language import Language

import nltk
from nltk.tokenize import BlanklineTokenizer

from handlers.output_generator import generate_output
from core.constants import constants
from core.config import Settings


class SpacyExtractor:
    """class SpacyExtractor encapsulates logic to pipe Records with an id and text body
    through a spacy model and return entities separated by Entity Type
    """

    def __init__(
        self, nlp: Language, input_id_col: str = "id", input_text_col: str = "text"
    ):
        """Initialize the SpacyExtractor pipeline.
        
        nlp (spacy.language.Language): pre-loaded spacy language model
        input_text_col (str): property on each document to run the model on
        input_id_col (str): property on each document to correlate with request

        RETURNS (EntityRecognizer): The newly constructed object.
        """
        self.nlp = nlp
        self.input_id_col = input_id_col
        self.input_text_col = input_text_col

    def _name_to_id(self, text: str):
        """Utility function to do a messy normalization of an entity name

        text (str): text to create "id" from
        """
        return "-".join([s.lower() for s in text.split()])

    def extract_entities(self, records: List[Dict[str, str]]):
        """Apply the pre-trained model to a batch of records
        
        records (list): The list of "document" dictionaries each with an
            `id` and `text` property
        
        RETURNS (list): List of responses containing the id of 
            the correlating document and a list of entities.
        """
        ids = (doc[self.input_id_col] for doc in records)
        texts = (doc[self.input_text_col] for doc in records)

        res = []

        for doc_id, spacy_doc in zip(ids, self.nlp.pipe(texts)):
            entities = {}
            for ent in spacy_doc.ents:
                ent_id = ent.kb_id
                if not ent_id:
                    ent_id = ent.ent_id
                if not ent_id:
                    ent_id = self._name_to_id(ent.text)

                if ent_id not in entities:
                    if ent.text.lower() == ent.text:
                        ent_name = ent.text.capitalize()
                    else:
                        ent_name = ent.text
                    entities[ent_id] = {
                        "name": ent_name,
                        "label": ent.label_,
                        "matches": [],
                    }
                entities[ent_id]["matches"].append(
                    {"start": ent.start_char, "end": ent.end_char, "text": ent.text}
                )

            res.append({"id": doc_id, "entities": list(entities.values())})
        return res
    
    def predict_nlp_component(self, component_type: str, input_text: str):

        doc = self.nlp(input_text)

        if component_type.lower() == constants.COMPONENT_TYPE_TOKENIZE.lower():
            token_data = []

            for token in doc:
                token_data.append(token.text)

            return generate_output(True, data=token_data)
        
        elif component_type.lower() == constants.COMPONENT_TYPE_POS.lower():
            pos_data = []
            pos_token: typing.Dict[str, typing.Any] = {}

            for token in doc:
                pos_token['text'] = str(token.text)
                pos_token['pos'] = str(token.pos_)

                pos_data.append(pos_token.copy())

            return generate_output(True, data=pos_data)
        
        elif component_type.lower() == constants.COMPONENT_TYPE_LEMMATIZER.lower():
            lemma_data = []
            lemma_token: typing.Dict[str, typing.Any] = {}

            for token in doc:
                lemma_token['text'] = str(token.text)
                lemma_token['lemma'] = str(token.lemma_)

                lemma_data.append(lemma_token.copy())

            return generate_output(True, data=lemma_data)
        
        elif component_type.lower() == constants.COMPONENT_TYPE_MORPHOLOGY.lower():
            morphology_data = []
            morphology_token: typing.Dict[str, typing.Any] = {}

            for token in doc:
                morphology_token['text'] = str(token.text)
                morphology_token['morphology'] = str(token.morph)

                morphology_data.append(morphology_token.copy())
                
            return generate_output(True, data=morphology_data)
        
        elif component_type.lower() == constants.COMPONENT_TYPE_DEPENDENCY_PARSER.lower():
            dep_data = []
            dep_token: typing.Dict[str, typing.Any] = {}

            for token in doc:
                dep_token['text'] = str(token.text)
                dep_token['head'] = str(token.head.text)
                dep_token['head_pos'] = str(token.head.pos_)
                dep_token['children'] = [str(child) for child in token.children]

                dep_data.append(dep_token.copy())
                
            return generate_output(True, data=dep_data)
        
        elif component_type.lower() == constants.COMPONENT_TYPE_NER.lower():
            ner_data = []
            ner_token: typing.Dict[str, typing.Any] = {}

            for ent in doc.ents:
                ner_token['text'] = str(ent.text)
                ner_token['start'] = str(ent.start_char)
                ner_token['end'] = str(ent.end_char)
                ner_token['entity'] = str(ent.label_)

                ner_data.append(ner_token.copy())
                
            return generate_output(True, data=ner_data)
        
        elif component_type.lower() == constants.COMPONENT_TYPE_NORMALIZERS.lower():
            norm_data = []
            norm_token: typing.Dict[str, typing.Any] = {}

            for token in doc:
                norm_token['text'] = str(token.text)
                norm_token['norm'] = str(token.norm_)

                norm_data.append(norm_token.copy())
                
            return generate_output(True, data=norm_data)
        
    def deep_linguistic_analysis(self, input_text: str) -> typing.Dict[str, typing.Any]:
        
        deep_linguistic_struc: typing.Dict[str, typing.Any] = {}  # deep_linguistic_struc

        nlp = spacy.load(Settings.SPACY_MODEL)
        doc = nlp(input_text)
        assert doc.has_annotation("SENT_START")

        paragraphs_line_struc: typing.Dict[str, typing.Any] = {}  #  paragraphs_line_struc
        sentence_line_struc: typing.Dict[str, typing.Any] = {}  #  sentence_line_struc
        token_line_struc: typing.Dict[str, typing.Any] = {}  #  sentence_line_struc
        dependency_line_struc: typing.Dict[str, typing.Any] = {}  #  dependency_line_struc

        deep_linguistic_struc['paragraphs'] = []
        deep_linguistic_struc['sentences'] = []
        deep_linguistic_struc['token'] = []

        tokenized_paragraphs = BlanklineTokenizer().tokenize(input_text)

        for paragraph in tokenized_paragraphs:
            paragraphs_line_struc['paragraph'] = str(paragraph)
            deep_linguistic_struc['paragraphs'].append(paragraphs_line_struc.copy())

        for sentence in doc.sents:
            sentence_line_struc['phrase'] = str(sentence)
            sentence_line_struc['start'] = int(sentence.start_char)
            sentence_line_struc['end'] = int(sentence.end_char)
            deep_linguistic_struc['sentences'].append(sentence_line_struc.copy())

        for token in doc:
            token_line_struc['name'] = str(token.text)
            token_line_struc['index'] = str(token.i)
            token_line_struc['pos'] = str(token.pos_)
            token_line_struc['lemma'] = str(token.lemma_)
            token_line_struc['norm'] = str(token.norm_)
            token_line_struc['tag'] = str(token.tag_)
            token_line_struc['shape'] = str(token.shape_)
            token_line_struc['is_alpha'] = str(token.is_alpha)
            token_line_struc['is_stop'] = str(token.is_stop)
            token_line_struc['morphology'] = str(token.morph)

            dependency_line_struc['dep'] = str(token.dep_)
            dependency_line_struc['head'] = str(token.head.text)
            dependency_line_struc['head_pos'] = str(token.head.pos_)
            dependency_line_struc['children'] = [str(child) for child in token.children]
            token_line_struc['dependency'] = dependency_line_struc.copy()

            deep_linguistic_struc['token'].append(token_line_struc.copy())

        return deep_linguistic_struc
    
    def entity_recognition(self, input_text: str) -> typing.Dict[str, typing.Any]:
        
        entity_recognition_struc: typing.Dict[str, typing.Any] = {}  # deep_linguistic_struc
        entity_line_struc: typing.Dict[str, typing.Any] = {}  #  paragraphs_line_struc

        nlp = spacy.load(Settings.SPACY_MODEL)
        doc = nlp(input_text)
        assert doc.has_annotation("SENT_START")

        entity_recognition_struc['entities'] = []

        for ent in doc.ents:
            entity_line_struc['entity'] = str(ent.text)
            entity_line_struc['category'] = str(ent.label_)
            entity_line_struc['start'] = str(ent.start_char)
            entity_line_struc['end'] = str(ent.end_char)
            
            entity_recognition_struc['entities'].append(entity_line_struc.copy())
            
        return entity_recognition_struc



from dataclasses import dataclass
import os

@dataclass
class Settings:

    LOG_FILE_PATH : str=os.path.join('app.log')
    LOG_FILE_FORMAT: str='%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    SPACY_MODEL: str='en_core_web_lg'

    _CUSTOM_SUB_CSV_FILE_PATH : str=os.path.join('data\\raw',"custom_substitutions.csv")

settings = Settings()


import json

class CustomJSONError(Exception):
    def __init__(self, error_code, message):
        self.error_code = error_code
        self.message = message
        super().__init__(self.message)

    def to_json(self):
        if hasattr(self, 'error_code'):
            error_dict = {
                "succeeded": False,
                "error": {
                    "error_code": self.error_code,
                    "message": self.message
                }
            }
            return json.dumps(error_dict)
        else:
            result_dict = {
                "succeeded": True,
                "data": self.message
            }
            return json.dumps(result_dict)
